import enum


class TipoArquivo(enum.Enum):
    TXT = "T"
    DBF = "D"


class Arquivo(enum.Enum):
    ARQUIVO_DE_APLICACAO_RESGATE = "002"
    ARQUIVO_DE_COTACOES = "003"
    CADASTRO_DE_FUNDOS = "004"
    OPERACOES_EM_SER_DO_PASSIVO = "005"
    CADASTRO_DE_GESTORES = "006"
    CADASTRO_DE_GRUPOS = "007"
    CADASTRO_DE_USUARIOS = "008"
    CADASTRO_DE_ASSISTENTES = "009"
    ARQUIVO_DE_OPERACOES_EM_SER_MENSAL = "010"
    POSICAO_DE_FUNDOS_DE_PRIVATIZACAO = "011"
    ARQUIVO_DE_PERFORMANCE = "012"
    ARQUIVO_CAC_TOTAL = "013"
    ARQUIVO_ARA_PARCIAL = "014"
    ARQUIVO_ARA_TOTAL = "015"
    ARQUIVO_RAF_PARCIAL = "016"
    ARQUIVO_RES_PARCIAL = "017"
    ARQUIVO_DE_CONTROLE_DE_ARQUIVOS_GERADOS = "018"
    ARQUIVO_DE_CADASTRO_DE_COTITULARES = "019"
    OPERACOES_EM_SER_DO_PASSIVO_D1 = "021"
    OPERACOES_EM_SER_DO_PASSIVO_D2 = "022"
    ARQUIVO_DE_EXTRATO_COMPENSACAO_DE_PREJUIZO = "023"
    RECOLHIMENTO_DE_TRIBUTOS_POR_PERIODO = "024"
    POSICAO_RESULTANTE_DE_RENDIMENTOS_A_COMPENSAR = "025"
    FECHAMENTO_DE_COTAS_EM_SER_D0 = "026"
    FECHAMENTO_DE_COTAS_EM_SER_D1 = "027"
    FECHAMENTO_DE_COTAS_EM_SER_D2 = "028"
    APLICACOES_EM_ABERTO_COTAS_DE_ABERTURA = "029"
    APLICACOES_EM_ABERTO_COTAS_DE_ABERTURA_D1 = "030"
    APLICACOES_EM_ABERTO_COTAS_DE_ABERTURA_D2 = "031"
    APLICACOES_EM_ABERTO_COTAS_DE_FECHAMENTO = "032"
    APLICACOES_EM_ABERTO_COTAS_DE_FECHAMENTO_D1 = "033"
    APLICACOES_EM_ABERTO_COTAS_DE_FECHAMENTO_D2 = "034"
    ARQUIVO_DE_MOVIMENTACOES_EFETIVADAS = "035"
    ARQUIVO_DE_MOVIMENTACOES_PENDENTES = "036"
    CADASTRO_DE_COTISTAS_EXPANDIDO = "037"
    ARQUIVO_ARA_CERTIFICADO_PARCIAL = "038"
    ARQUIVO_ARA_CERTIFICADO_TOTAL = "039"
    ARQUIVO_DE_CONTAS_DE_CREDITO_INVESTIMENTO = "041"
    ARQUIVO_DE_BLOQUEIOS = "050"
    ARQUIVO_DE_CLASSIFICACAO_DE_COTISTAS = "051"
    EXTRATO_COMPENSACAO_DE_PREJUIZO_FECHAMENTO = "053"
    RECOLHIMENTO_DE_TRIBUTOS_POR_DECENDIO = "056"
    OPERACOES_EM_SER_REGERADAS = "057"
    APLICACOES_EM_ABERTO_REGERADAS = "058"
    ARQUIVO_DE_SALDOS_COTA_DE_ABERTURA_D0 = "062"
    ARQUIVO_DE_SALDOS_COTA_DE_ABERTURA_D_1 = "063"
    ARQUIVO_DE_SALDOS_COTA_DE_ABERTURA_D_2 = "064"
    ARQUIVO_DE_SALDOS_COTA_DE_FECHAMENTO_D0 = "065"
    ARQUIVO_DE_SALDOS_COTA_DE_FECHAMENTO_D_1 = "066"
    ARQUIVO_DE_SALDOS_COTA_DE_FECHAMENTO_D_2 = "067"
    ARQUIVO_DE_SALDOS_MENSAL = "068"
    PESSOAS_VINCULADAS_DO_COTISTA = "069"
    CADASTRO_DE_CONTAS_DE_CREDITO_COMPLETO_D_1 = "073"
    MOVIMENTO_DIARIO_D_1 = "074"
    OPERACOES_EMSER_DO_PASSIVO_SIMULADA = "075"
    ARQUIVO_DE_RESGATES_AGENDADOS = "076"
    RELACAO_DE_DOCUMENTOS_PENDENTES = "077"
    ARQUIVO_DE_EXTRATO_COMPENSACAO_DE_PREJUIZO_ = "078"
    EXTRATO_COMPENSACAO_DE_PREJUIZO_CPF_FECHA = "079"
    MOVIMENTO_HISTORICO_DE_PREJUIZO_CPF = "080"
    POSICAO_RESULTANTE_DE_RENDIMENTOS_A_COMPENSAR_ = "081"
    ARQUIVO_DE_OPERACOES_DE_EXTRATO_REPROCESSADO = "082"
    CADASTRO_DE_COTISTAS_INTRAG = "083"
    PES_VINCULADAS_COTISTA_INTRAG = "084"
    CADASTRO_CO_TITULARES_INTRAG = "085"
    ACOMP_GIIN_PENDENTE_FATCA = "086"
    RISK_POINT_INTRAG = "089"
    ENVIO_DE_CORRESPONDENCIA = "091"
    E_FINANCEIRA_SALDOS = "092"
    E_FINANCEIRA_MOVIMENTOS = "093"
    E_FINANCEIRA_PAGAMENTOS = "094"
    MEMORIA_DE_CALCULO_AMORTIZACAO = "095"
    ARQUIVO_DE_SALDOS_AMORTIZADOS = "096"
    ISS_INFORMACAO_MENSAL_MERCADO = "097"
    CONSULTA_DE_MOVIMENTOS_CETIP = "098"
    CONSULTA_DE_CONTAS_CETIP = "099"
    POSICAO_COTISTA_DISTRIBUIDOR = "100"
    MOVIMENTOS_SIMULADOS_D0 = "101"
    INCONSISTENCIAS_CADASTRAIS = "102"
    ARQUIVO_TCR_GERAL = "103"



from io import BufferedReader, BytesIO
import struct
import datetime
import os
import codecs

# Check for optional dependencies.
import pandas as pd


class DbfBase(object):
    """
    Base class for DBF file processing objects.

    Do not instantiate this class. This provides some of the common functions
    for other subclasses.
    """

    def _chunker(self, chunksize):
        """Return a list of chunk ints from given chunksize.

        Parameters
        ----------
        chunksize : int
            The maximum chunk size

        Returns
        -------
        list of ints
            A list of chunks necessary to break up a given file. These will
            all be equal to `chunksize`, except for the last value, which is
            the remainder (<= `chunksize).
        """
        num = self.numrec // chunksize
        # Chunksize bigger than numrec
        if num == 0:
            return [
                self.numrec,
            ]
        else:
            chunks = [
                chunksize,
            ] * num
            remain = self.numrec % chunksize
            if remain != 0:
                chunks.append(remain)
            return chunks

    def _na_set(self, na):
        """Set the value used for missing/bad data.

        Parameters
        ----------
        na : various types accepted
            The value that will be used to replace missing or malformed
            entries. Right now this accepts pretty much anything, and that
            value will be used as a replacement. (May not do what you expect.)
            However, the strings 'na' or 'nan' (case insensitive) will insert
            float('nan'), the string 'none' (case insensitive) or will insert
            the Python object `None`.  Float/int columns are always
            float('nan') regardless of this setting.
        """
        if na.lower() == "none":
            self._na = None
        elif na.lower() in ("na", "nan"):
            self._na = float("nan")
        else:
            self._na = na

    def mem(self, chunksize=None):
        """Print the memory usage for processing the DBF File.

        Parameters
        ----------
        chunksize : int, optional
            The maximum chunk size that will be used to process this file.

        Notes
        -----
        This method will print the maximum amount of RAM that will be
        necessary to process and load the DBF file. (This is ~2x the file
        size.) However, if the optional chunksize is passed, this function
        will also print memory usage per chunk as well, which can be useful
        for efficiently chunking and processing a file.
        """
        if chunksize:
            if chunksize > self.numrec:
                print("Chunksize larger than number of recs.")
                print("Chunksize set to {:d}.".format(self.numrec))
            else:
                smallmem = 2.0 * (self.fmtsiz * chunksize / 1024**2)
                chkout = "Each chunk will require {:.4g} MB of RAM."
                print(chkout.format(smallmem))
        memory = 2.0 * (self.fmtsiz * self.numrec / 1024**2)
        out = "This total process would require more than {:.4g} MB of RAM."
        print(out.format(memory))

    def to_dataframe(self, chunksize=None, na="nan"):
        """Return the DBF contents as a DataFrame.

        Parameters
        ----------
        chunksize : int, optional
            Maximum number of records to process at any given time. If 'None'
            (defalut), process all records.

        na : various types accepted, optional
            The value that will be used to replace missing or malformed
            entries. Right now this accepts pretty much anything, and that
            value will be used as a replacement. (May not do what you expect.)
            However, the strings 'na' or 'nan' (case insensitive) will insert
            float('nan'), the string 'none' (case insensitive) or will insert
            the Python object `None`. Default for DataFrame is NaN ('nan');
            however, float/int columns are always float('nan')

        Returns
        -------
        DataFrame (chunksize == None)
            The DBF file contents as a Pandas DataFrame

        Generator (chunksize != None)
            This generator returns DataFrames with the maximum number of
            records equal to chunksize. (May be less)

        Notes
        -----
        This method requires Pandas >= 0.15.2.
        """
        self._na_set(na)
        if not chunksize:
            # _get_recs is a generator, convert to list for DataFrame
            results = list(self._get_recs())
            df = pd.DataFrame(results, columns=self.columns)
            del results  # Free up the memory? If GC works properly
            return df
        else:
            # Return a generator function instead
            return self._df_chunks(chunksize)

    def _df_chunks(self, chunksize):
        """A DataFrame chunk generator.

        See `to_dataframe`.
        """
        chunks = self._chunker(chunksize)
        # Keep track of the index, otherwise every DataFrame will be indexed
        # starting at 0
        idx = 0
        for chunk in chunks:
            results = list(self._get_recs(chunk=chunk))
            num = len(results)  # Avoids skipped records problem
            df = pd.DataFrame(
                results, columns=self.columns, index=range(idx, idx + num)
            )
            idx += num
            del results
            yield df


class Dbf5(DbfBase):
    """
    DBF version 5 file processing object.

    This class defines the methods necessary for reading the header and
    records from a version 5 DBF file.  Much of this code is based on an
    `ActiveState DBF example`_, which only worked for Python2.

    .. ActiveState DBF example: http://code.activestate.com/recipes/
            362715-dbf-reader-and-writer/

    Parameters
    ----------

    dbf : string
        The name (with optional path) of the DBF file.

    codec : string, optional
        The codec to use when decoding text-based records. The default is
        'utf-8'. See Python's `codec` standard lib module for other options.

    Attributes
    ----------

    dbf : string
        The input file name.

    f : file object
        The opened DBF file object

    numrec : int
        The number of records contained in this file.

    lenheader : int
        The length of the file header in bytes.

    numfields : int
        The number of data columns.

    fields : list of tuples
        Column descriptions as a tuple: (Name, Type, # of bytes).

    columns : list
        The names of the data columns.

    fmt : string
        The format string that is used to unpack each record from the file.

    fmtsiz : int
        The size of each record in bytes.
    """

    def __init__(self, dbf: bytes, codec="utf-8"):
        bytes_io = BytesIO(dbf)

        buffer_reader = BufferedReader(bytes_io)

        self._enc = codec
        self._esc = None
        # Reading as binary so bytes will always be returned
        self.f = buffer_reader

        self.numrec, self.lenheader = struct.unpack("<xxxxLH22x", self.f.read(32))
        self.numfields = (self.lenheader - 33) // 32

        # The first field is always a one byte deletion flag
        fields = [
            ("DeletionFlag", "C", 1),
        ]
        for fieldno in range(self.numfields):
            name, typ, size = struct.unpack("<11sc4xB15x", self.f.read(32))
            # eliminate NUL bytes from name string
            name = name.strip(b"\x00")
            fields.append((name.decode(self._enc), typ.decode(self._enc), size))
        self.fields = fields
        # Get the names only for DataFrame generation, skip delete flag
        self.columns = [f[0] for f in self.fields[1:]]

        terminator = self.f.read(1)
        assert terminator == b"\r"

        # Make a format string for extracting the data. In version 5 DBF, all
        # fields are some sort of structured string
        self.fmt = "".join(["{:d}s".format(fieldinfo[2]) for fieldinfo in self.fields])
        self.fmtsiz = struct.calcsize(self.fmt)

    def _get_recs(self, chunk=None):
        """Generator that returns individual records.

        Parameters
        ----------
        chunk : int, optional
            Number of records to return as a single chunk. Default 'None',
            which uses all records.
        """
        if chunk == None:
            chunk = self.numrec

        for i in range(chunk):
            # Extract a single record
            record = struct.unpack(self.fmt, self.f.read(self.fmtsiz))
            # If delete byte is not a space, record was deleted so skip
            if record[0] != b" ":
                continue

            # Save the column types for later
            self._dtypes = {}
            result = []
            for idx, value in enumerate(record):
                name, typ, size = self.fields[idx]
                if name == "DeletionFlag":
                    continue

                # String (character) types, remove excess white space
                if typ == "C":
                    if name not in self._dtypes:
                        self._dtypes[name] = "str"
                    value = value.strip()
                    # Convert empty strings to NaN
                    if value == b"":
                        value = self._na
                    else:
                        value = value.decode(self._enc)
                        # Escape quoted characters
                        if self._esc:
                            value = value.replace('"', self._esc + '"')

                # Numeric type. Stored as string
                elif typ == "N":
                    # A decimal should indicate a float
                    if b"." in value:
                        if name not in self._dtypes:
                            self._dtypes[name] = "float"
                        value = float(value)
                    # No decimal, probably an integer, but if that fails,
                    # probably NaN
                    else:
                        try:
                            value = int(value)
                            if name not in self._dtypes:
                                self._dtypes[name] = "int"
                        except:
                            # I changed this for SQL->Pandas conversion
                            # Otherwise floats were not showing up correctly
                            value = float("nan")

                # Date stores as string "YYYYMMDD", convert to datetime
                elif typ == "D":
                    try:
                        y, m, d = int(value[:4]), int(value[4:6]), int(value[6:8])
                        if name not in self._dtypes:
                            self._dtypes[name] = "date"
                    except:
                        value = self._na
                    else:
                        value = datetime.date(y, m, d)

                # Booleans can have multiple entry values
                elif typ == "L":
                    if name not in self._dtypes:
                        self._dtypes[name] = "bool"
                    if value in b"TyTt":
                        value = True
                    elif value in b"NnFf":
                        value = False
                    # '?' indicates an empty value, convert this to NaN
                    else:
                        value = self._na

                # Floating points are also stored as strings.
                elif typ == "F":
                    if name not in self._dtypes:
                        self._dtypes[name] = "float"
                    try:
                        value = float(value)
                    except:
                        value = float("nan")

                else:
                    err = 'Column type "{}" not yet supported.'
                    raise ValueError(err.format(value))

                result.append(value)
            yield result



from dataclasses import dataclass
from typing import List, Union, cast
import requests
import bs4
import re
import datetime
import pandas as pd
import xml.etree.ElementTree as ET
from intrag_sdk.passivo.file_types import TipoArquivo, Arquivo
import zipfile
import io

from intrag_sdk.simpledbf import Dbf5

DEFAULT_ENCODING = "ISO-8859-1"


class FileNotFound(Exception):
    pass


@dataclass
class Download:
    file_name: str
    data: Union[str, pd.DataFrame]


class ItauPassivo:
    headers = {
        "cache-control": "max-age=0",
        "content-type": "application/x-www-form-urlencoded",
    }

    def __init__(self, echo=False):
        self.base_url = "https://www.itaucustodia.com.br/Passivo"
        self.cookies = {}
        self.echo = echo

    def endpoint(self, path: str):
        return f"{self.base_url}{path}"

    def post(self, endpoint: str, data: dict = dict()):
        if self.echo:
            print(f"POST {endpoint}")
            print(data)

        form_data = "&".join([f"{key}={value}" for key, value in data.items()])

        res = requests.post(
            self.endpoint(endpoint),
            headers=self.headers,
            cookies=self.cookies,
            data=form_data,
        )

        self.cookies = {**self.cookies, **res.cookies.get_dict()}

        return res

    def __fetch_gestor_info(self):
        """Seta codigo do gestor e lista de fundos disponiveis"""
        res = self.post("/abreFiltroConsultaMovimentoFundoTotais.do")

        html = bs4.BeautifulSoup(
            res.content, "html.parser", from_encoding=DEFAULT_ENCODING
        )

        gestor = html.find("select", dict(name="codigoGestor")).find_all("option")[-1]

        codigo_gestor = gestor.attrs.get("value")
        nome_gestor = gestor.text

        self.nome_gestor = nome_gestor
        self.codigo_gestor = codigo_gestor

    def authenticate(self, user: str, password: str):
        res = self.post("/login.do", data={"ebusiness": user, "senha": password})

        if "logoff" not in res.content.decode(encoding=DEFAULT_ENCODING):
            raise Exception("Login Inválido")

        self.cookies = {"JSESSIONID": res.cookies.get("JSESSIONID")}
        self.__fetch_gestor_info()
        self.__fetch_funds_info()

    def __fetch_funds_info(self):
        try:
            downloads = self.download_de_arquivos(arquivo=Arquivo.CADASTRO_DE_FUNDOS)

            downloads = cast(List[Download], downloads)

            self.fundos = cast(pd.DataFrame, downloads[0].data)

        except FileNotFound:
            self.fundos = pd.DataFrame()

    def download_de_arquivos(
        self,
        arquivo: Arquivo,
        tipo_arquivo: TipoArquivo = TipoArquivo.DBF,
        data: datetime.date = datetime.date.today(),
        raw: bool = False,
    ) -> Union[List[Download], zipfile.ZipFile]:
        """
        Baixa arquivos do servidor e retorna uma lista de objetos Download ou um arquivo zip, dependendo do valor do parâmetro raw.

        Argumentos:
            tipo_arquivo (TipoArquivo): O tipo de arquivo a ser baixado.
            arquivo (Arquivo): O arquivo a ser baixado.
            data (datetime.date, opcional): A data do arquivo a ser baixado. O padrão é a data de hoje.
            raw (bool, opcional): Se True, retorna um arquivo zip. Se False, retorna uma lista de objetos Download. O padrão é False.

        Retorna:
            Union[List[Download], zipfile.ZipFile]: Uma lista de objetos Download ou um arquivo zip, dependendo do valor do parâmetro raw.
        """

        date_str = data.strftime("%d%m%Y")

        self.post(
            "/listarOpcoesArquivosDownloadArquivos.do",
            data={
                "codigoGestor": self.codigo_gestor,
                "tipoArquivo": tipo_arquivo.value,
                "data": date_str,
            },
        )

        self.post(
            "/processarDownloadArquivosAjax.do",
            data={
                "codigoGestor": self.codigo_gestor,
                "tipoArquivo": tipo_arquivo.value,
                "numeroArquivo": arquivo.value,
            },
        )

        res = self.post(
            "/EfetuarDownloadArquivosListaServlet",
            data={
                "checkArquivos": arquivo.value,
                "numerosArquivosSelecionados": arquivo.value,
            },
        )

        if "zip" not in res.headers["Content-Type"]:
            raise FileNotFound("Arquivo não encontrado")

        archive = zipfile.ZipFile(io.BytesIO(res.content), "r")

        if raw:
            return archive

        downloads = []

        for file in archive.filelist:
            raw_file_data = archive.read(file)

            file_data = None

            if tipo_arquivo == TipoArquivo.DBF:
                file_data = cast(pd.DataFrame, Dbf5(raw_file_data).to_dataframe())

            if tipo_arquivo == TipoArquivo.TXT:
                file_data = raw_file_data.decode("utf-8")

            if file_data is not None:
                downloads.append(Download(file_name=file.filename, data=file_data))

        return downloads

    def posicao_cotistas(self, codigo_fundo: str):
        """
        codigo_fundo (str): Pode ser encontrado em 'fundos'

        ```
        client = ItauPassivo()
        client.fundos["CDFDO"].
        ```
        """

        data = {
            "codigoGestor": self.codigo_gestor,
            "codigoFundo": codigo_fundo,
        }

        res = self.post("/consultarCotistasFundo.do", data=data)

        html = make_soup(res)

        tables = html.find("div", {"id": "listaDados"}).find_all("table")

        data = list(map(parse_table, tables))

        return pd.DataFrame(data)

    def movimentos_do_dia(
        self,
    ):
        data = {
            "codigoGestor": self.codigo_gestor,
        }

        res = self.post("/consultarMovimentoDia.do", data=data)

        html = make_soup(res)

        movimentos_dia = html.find("span", string="Movimentos do Dia")

        if movimentos_dia is None:
            return pd.DataFrame()

        tables = movimentos_dia.find_next_siblings("table")[:-2]

        data = list(map(parse_table, tables))

        return pd.DataFrame(data)


def make_soup(res, encoding=DEFAULT_ENCODING):
    return bs4.BeautifulSoup(res.content, "html.parser", from_encoding=encoding)


def parse_table(table):
    tds = table.find_all("td")

    td_tuples = [(tds[i], tds[i + 1]) for i in range(0, len(tds), 2)]

    def parse_key(td):
        return td.text.strip()[:-1]

    def parse_value(td):
        text = td.text.strip()

        _text = text.replace(".", "").replace(",", "")

        if _text.isnumeric():
            return float(text.replace(".", "").replace(",", "."))

        regex = r"(\d{2})/(\d{2})/(\d{4})"

        if re.match(regex, text):
            return datetime.datetime.strptime(text, "%d/%m/%Y").date()

        return text

    return {parse_key(key): parse_value(value) for key, value in td_tuples}



import datetime
from typing import List
from zipfile import ZipFile
from intrag_sdk import ItauPassivo, TipoArquivo, Arquivo
import pandas as pd
from intrag_sdk.passivo.client import Download
from intrag_sdk.simpledbf import Dbf5

user = "rps.op04"
password = "996329"

api = ItauPassivo(echo=True)
api.authenticate(user=user, password=password)

# Propriedades disponíveis
nome_gestor: str = api.nome_gestor
codigo_gestor: str = api.codigo_gestor
fundos: pd.DataFrame = api.fundos


# Download de arquivos
downloads: List[Download] | ZipFile = api.download_de_arquivos(
    Arquivo.ARQUIVO_DE_PERFORMANCE, data=datetime.date(2023, 10, 6)
)

# Movimentações do dia
movimentacoes: pd.DataFrame = api.movimentos_do_dia()

# Posição de cotistas
codigo_fundo: str = fundos["CDFDO"][0]

posicoes: pd.DataFrame = api.posicao_cotistas(codigo_fundo=codigo_fundo)
